import jsonschema.validators


def _set_defaults(validator, properties, instance, schema):
    for property, subschema in properties.items():
        if "default" in subschema and property not in instance:
            instance[property] = subschema["default"]


DefaultSetter = jsonschema.validators.create(
    meta_schema={}, validators={"properties": _set_defaults},
)


def extend(validator_cls):
    """
    Extend the given :class:`jsonschema.IValidator` with the Seep layer.

    """

    Validator = jsonschema.validators.extend(
        validator_cls, {
            "properties" : _properties_with_defaults(validator_cls),
        }
    )

    class Blueprinter(Validator):
        def instantiate(self, data):
            self.validate(data)
            return data

    return Blueprinter


def instantiate(data, blueprint):
    """
    Instantiate the given data using the blueprinter.

    :argument blueprint: a blueprint (JSON Schema with Seep properties)

    """

    Validator = jsonschema.validators.validator_for(blueprint)
    blueprinter = extend(Validator)(blueprint)
    return blueprinter.instantiate(data)


def _properties_with_defaults(validator_cls):
    def properties_with_defaults(validator, properties, instance, schema):
        for error in validator_cls.VALIDATORS["properties"](
            validator, properties, instance, schema
        ):
            yield error

        subschemas = [(instance, schema)]
        while subschemas:
            subinstance, subschema = subschemas.pop()
            DefaultSetter(subschema).validate(subinstance)
            subschemas.extend(
                (subinstance.setdefault(property, {}), subsubschema)
                for property, subsubschema in
                subschema.get("properties", {}).items()
            )

    return properties_with_defaults



from unittest import TestCase

import jsonschema
import seep.core


class TestInstantiate(TestCase):
    def test_it_sets_defaults(self):
        data = {}
        schema = {"properties" : {"foo" : {"default" : 12}}}
        seep.core.instantiate(data, schema)
        self.assertEqual(data, {"foo" : 12})

    def test_it_sets_nested_defaults(self):
        data = {}
        schema = {
            "properties" : {
                "foo" : {
                    "default" : {},
                    "properties" : {"bar" : {"default" : []}},
                }
            }
        }
        seep.core.instantiate(data, schema)
        self.assertEqual(data, {"foo" : {"bar" : []}})

    def test_it_sets_multiple_nested_defaults(self):
        data = {}
        schema = {
            "properties" : {
                "bar" : {"default" : 123},
                "foo" : {
                    "default" : {},
                    "properties" : {
                        "bar" : {
                            "properties": {"baz" : {"default" : []}},
                        },
                    },
                },
            }
        }
        seep.core.instantiate(data, schema)
        self.assertEqual(data, {"bar": 123, "foo" : {"bar" : {"baz" : []}}})

    def test_identity_instantiate(self):
        data = {"foo" : 12}
        schema = {"properties" : {"foo" : {}}}
        seep.core.instantiate(data, schema)
        self.assertEqual(data, {"foo" : 12})

    def test_validation_errors_are_still_errors(self):
        with self.assertRaises(jsonschema.ValidationError):
            seep.core.instantiate("foo", {"type" : "integer"})



import os

from setuptools import find_packages, setup


with open(os.path.join(os.path.dirname(__file__), "README.rst")) as readme:
    long_description = readme.read()

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy"
]

setup(
    name="seep",
    packages=find_packages(),
    install_requires=["jsonschema"],
    setup_requires=["vcversioner"],
    author="Julian Berman",
    author_email="Julian@GrayVines.com",
    classifiers=classifiers,
    description="A data coersion language built on top of jsonschema",
    license="MIT",
    long_description=long_description,
    url="https://github.com/Julian/Seep",
    vcversioner={"version_module_paths" : ["seep/_version.py"]},
)
